/* extension.js
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: GPL-2.0-or-later
 */

/* exported init */
const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();

const { St, Clutter, Graphene, Pango } = imports.gi;
const Calendar = imports.ui.calendar;
const MessageTray = imports.ui.messageTray;
const MessageList = imports.ui.messageList;
const PopupMenu = imports.ui.popupMenu;
const Main = imports.ui.main;
const Gettext = imports.gettext.domain(Me.metadata['gettext-domain']);
const _ = Gettext.gettext;
let _extension;

function getMode() {
    return _extension.expandMode;
}

function setExpandMode(mode) {
    _extension.expandMode = mode;
}

function expandMessages() {
    Main.panel.statusArea.dateMenu._messageList._notificationSection._list.get_children().forEach((m) => {
        if (m.child.forceExpansion) {
            m.child.expand(true);
        }
    });
}

function initiateLayout(message) {
    // Create a new layout
    message.pangoLayout = message.bodyLabel.clutter_text.create_pango_layout(message.bodyLabel._text);
    message.pangoLayout.set_wrap(Pango.WrapMode.WORD_CHAR);
    // Height of pango
    const heightPango = message._bodyStack.get_height() * Pango.SCALE;
    // Hardcoding the needed difference of actual width
    // This is the offset. The width of the message list is allocated, however we need to correct by an offset that is 
    // based on resolution
    let offset = 112;
    if (message.get_width() === 47.5) {
        offset = 130;
    }
    else if (message.get_width() === 39.5) {
        offset = 48;
    }
    // Calculated width
    const widthPango = Main.panel.statusArea.dateMenu._messageList.get_width() * Pango.SCALE - offset * Pango.SCALE;
    message.pangoLayout.set_height(heightPango);
    message.pangoLayout.set_width(widthPango);
}

function _onNotificationAdded(source, notification) {
    // The important line of the extension is this one. Here, usually a Object of the class NotificationMessage would be created.
    // However, by using NotificationBanner, access to more functions like expand and unexpand is possible.
    let message = notification.createBanner();//new MessageTray.NotificationBanner(notification);
    // NotificationBanners are too wide for the notification tray, so they have to be adjusted properly
    message.set_width(31.5);
    message.setSecondaryActor(new Calendar.TimeLabel(notification.datetime));
    let isUrgent = notification.urgency == MessageTray.Urgency.CRITICAL;

    let updatedId = notification.connect('updated', () => {
        message.setSecondaryActor(new Calendar.TimeLabel(notification.datetime));
        this.moveMessage(message, isUrgent ? 0 : this._nUrgent, this.mapped);
    });

    if (isUrgent) {
        // Keep track of urgent notifications to keep them on top
        this._nUrgent++;
    } else if (this.mapped) {
        // Only acknowledge non-urgent notifications in case it
        // has important actions that are inaccessible when not
        // shown as banner
        notification.acknowledged = true;
    }
    let index = isUrgent ? 0 : this._nUrgent;

    let destroyId = notification.connect('destroy', () => {
        notification.disconnect(destroyId);
        notification.disconnect(updatedId);
        if (isUrgent)
            this._nUrgent--;
        if (!expandButton) {
            // Nothing happens without a button.
        }
        else {
            // Make sure to also destroy the button
            expandButton.destroy();
        }
    });
    this.addMessageAtIndex(message, index, this.mapped);
    // The Icon we need
    let arrowIcon = PopupMenu.arrowIcon(St.Side.RIGHT);
    // there is no button yet, so it is null
    let expandButton = null;
    // Initiate Pango, so we can determine how many lines we need 
    initiateLayout(message);
    // Make sure we need to expand at all. Either the notification has a body and enough lines to expand, or it has actions
    if ((message._bodyText && (message.pangoLayout.get_line_count() > 1)) || notification.actions.length !== 0) {
        // Create the button
        expandButton = new St.Button({
            child: arrowIcon,
            y_align: Clutter.ActorAlign.START,
            y_expand: true,
            can_focus: true,
        });
        // There are 2 options: either all messages are auto expanded, or an arrow is added to the body.
        // However, without actions or with too few characters, there is no point to add the arrow to expand the notification.
        if (getMode() === true) {
            // Auto expand all
            // To force the expansion of the text
            message.forceExpansion = true;
        }
        else {
            // In able to align the arrow next to the description, a new BoxLayout has to be created
            let newBox = new St.BoxLayout({
                vertical: false,
                x_align: Clutter.ActorAlign.END,
                style_class: 'button-under-close'
            });
            // Re-creating the widget message._bodyStack and adding it to the new box as well as the expand button
            let newWidget = new St.Widget({ x_expand: true });
            newWidget.layout_manager = new MessageList.LabelExpanderLayout();
            let newBodyLabel = new MessageList.URLHighlighter('', false, message._useBodyMarkup);
            newBodyLabel.add_style_class_name('message-body');
            newWidget.add_actor(newBodyLabel);
            newBox.add(newWidget);
            newBox.add(expandButton);
            expandButton.x_align = Clutter.ActorAlign.END;
            expandButton.y_align = Clutter.ActorAlign.CENTER;
            message.child.get_first_child().get_children()[1].remove_actor(message._bodyStack);
            //log(message.child.get_first_child().get_children()[1].get_children());
            message._bodyStack = newWidget;
            message.bodyLabel = newBodyLabel;
            message.setBody(message._bodyText);
            // Inserting child at the correct position
            message.child.get_first_child().get_children()[1].insert_child_at_index(newBox, 2);
            // Connection button to signal
            expandButton.connect('clicked', () => {
                const originalYAlign = expandButton.y_align;
                if (message.expanded === true) {
                    // everything needed to unexpand. Make sure arrow is correctly aligned, change the icon of arrow, then unexpand
                    expandButton.y_align = originalYAlign;
                    expandButton.remove_actor(arrowIcon);
                    arrowIcon = PopupMenu.arrowIcon(St.Side.RIGHT);
                    expandButton.add_actor(arrowIcon);
                    message.clickedByButton = true;
                    message.keepActions = false;
                    message.unexpand(true);
                }
                else {
                    // Make sure to keep the arrow aligned at the top of the notification, so it does not move
                    if (originalYAlign === Clutter.ActorAlign.CENTER || originalYAlign === Clutter.ActorAlign.END) {
                        expandButton.y_align = Clutter.ActorAlign.START;
                    }
                    expandButton.remove_actor(arrowIcon);
                    arrowIcon = PopupMenu.arrowIcon(St.Side.BOTTOM);
                    expandButton.add_actor(arrowIcon);
                    message.clickedByButton = true;
                    message.expand(true);

                }
            });
        }
    }
}

function expand(animate) {
    this.oldBoxHeight = this.get_parent().get_height();
    this.expanded = true;
    this._actionBin.visible = this._actionBin.get_n_children() > 0;

    if (this._bodyStack.get_n_children() < 2) {
        this._expandedLabel = new MessageList.URLHighlighter(this._bodyText,
            true, this._useBodyMarkup);
        this.setExpandedBody(this._expandedLabel);
    }

    if (animate) {
        // Check if pango layout is already there, else create and assign
        if (!this.pangoLayout) {
            this.pangoLayout = this.bodyLabel.clutter_text.create_pango_layout(this.bodyLabel._text);
        }
        if (!this.clickedByButton && !this.forceExpansion) {
            // Usual animation
            this._bodyStack.ease_property('@layout.expansion', 1, {
                progress_mode: Clutter.AnimationMode.EASE_OUT_QUAD,
                duration: MessageTray.ANIMATION_TIME,
            });
        }
        else if (this.clickedByButton || this.forceExpansion) {
            let fix = 0;
            if (this._actionBin.get_height() !== 0) {
                // When expanding a notif for the second time, a bug exists that resets the height
                // of the actionBin. The height is saved when to this.actionBinHeight when inserting the message to notification list
                this._actionBin.set_height(this.actionBinHeight);
            }
            if (!this._bodyText) {
                // There is a weird issue with notifications without body, that ends up cutting a small part of the actionBin. This is a fix.
                fix = this._bodyStack.get_height();
            }
            else if (!this.nLinesSet && this.clickedByButton) {
                // The layout is already initiated, but we correct it by real width and heights
                this.pangoLayout.set_height((this.bodyLabel.content_box.get_height()) * Pango.SCALE);
                // The magic number 3 serves as an offset
                this.pangoLayout.set_width((this.get_parent().content_box.get_width() - 3) * Pango.SCALE);
                this._bodyStack.layout_manager._expandLines = this.pangoLayout.get_line_count();
                // Only once
                this.nLinesSet = true;
            }
            else if (!this.nLinesSet && this.forceExpansion) {
                initiateLayout(this);
                this._bodyStack.layout_manager._expandLines = this.pangoLayout.get_line_count();
                this.nLinesSet = true;
            }
            else {
                this._bodyStack.layout_manager._expandLines = this.pangoLayout.get_line_count();
            }
            this._bodyStack.layout_manager.expansion = 1;
            this.get_parent().set_height(this.get_height() + fix);
            this.get_parent().layout_manager.layout_changed();
        }

        this._actionBin.scale_y = 0;
        this._actionBin.ease({
            scale_y: 1,
            duration: MessageTray.ANIMATION_TIME,
            mode: Clutter.AnimationMode.EASE_OUT_QUAD,
        });

    } else {
        this._bodyStack.layout_manager.expansion = 1;
        this._actionBin.scale_y = 1;
    }
    this.emit('expanded');
}

function unexpand(animate) {
    if (animate) {
        this._bodyStack.ease_property('@layout.expansion', 0, {
            progress_mode: Clutter.AnimationMode.EASE_OUT_QUAD,
            duration: MessageTray.ANIMATION_TIME,
        });
        if (!this.keepActions) {
            this._actionBin.ease({
                scale_y: 0,
                duration: MessageTray.ANIMATION_TIME,
                mode: Clutter.AnimationMode.EASE_OUT_QUAD,
                onComplete: () => {
                    this._actionBin.hide();
                    this.expanded = false;
                },
            });
        } else {
            this.expanded = false;
        }
        if (this.clickedByButton) {
            this._bodyStack.layout_manager.expansion = 0;
            this._actionBin.scale_y = 0;
            this.get_parent().set_height(this.oldBoxHeight);
            this.get_parent().layout_manager.layout_changed();
            this.expanded = false;
        }
    } else {
        this._bodyStack.layout_manager.expansion = 0;
        if (!this.keepActions) this._actionBin.scale_y = 0;
        this.expanded = false;
    }
    this.emit('unexpanded');
}

function addMessageAtIndex(message, index, animate) {
    if (this._messages.includes(message))
        throw new Error('Message was already added previously');
    // Adjusting the listItem Bin to be different than usual, for testing
    let listItem = new St.Bin({
        child: message,
        reactive: true,
        y_expand: true,
        x_expand: true,
        pivot_point: new Graphene.Point({ x: .5, y: .5 }),
        layout_manager: new Clutter.BinLayout()
    });
    listItem._connectionsIds = [];

    listItem._connectionsIds.push(message.connect('key-focus-in',
        this._onKeyFocusIn.bind(this)));
    listItem._connectionsIds.push(message.connect('close', () => {
        this.removeMessage(message, true);
    }));
    listItem._connectionsIds.push(message.connect('destroy', () => {
        listItem._connectionsIds.forEach(id => message.disconnect(id));
        listItem.destroy();
    }));

    this._list.insert_child_at_index(listItem, index);
    if (animate) {
        listItem.set({ scale_x: 0, scale_y: 0 });
        listItem.ease({
            scale_x: 1,
            scale_y: 1,
            duration: MessageList.MESSAGE_ANIMATION_TIME,
            mode: Clutter.AnimationMode.EASE_OUT_QUAD,
            y_expand: true
        });
    }
    // Saving the height of the actionBin prematurely since there exists a bug
    message.actionBinHeight = message._actionBin.get_height();
}
// Messages are expanded as soon as notification list is opened
function _onOpenStateChanged(menu, open) {
    _originalOnOpenStateChanged.apply(this, [menu, open]);
    if (Main.panel.statusArea.dateMenu._messageList._notificationSection._list.length !== 0) {
        expandMessages();
    }

}

// These variables are used to store original functions
let _originalOnNotificationAdded;
let originalExpand;
let originalUnexpand;
let originalAddMessageAtIndex;
let _originalOnOpenStateChanged;
let settingsHandler = null;
let _originalOnMenuKeyPress;
class Extension {
    constructor(uuid) {
        this._uuid = uuid;
        ExtensionUtils.initTranslations(Me.metadata['gettext-domain']);
    }
    enable() {
        // Enable settings
        this.settings = ExtensionUtils.getSettings(Me.metadata['settings-schema']);
        // Get original functions
        _originalOnNotificationAdded = Calendar.NotificationSection.prototype._onNotificationAdded;
        originalExpand = MessageList.Message.prototype.expand;
        originalUnexpand = MessageList.Message.prototype.unexpand;
        originalAddMessageAtIndex = MessageList.Message.prototype.addMessageAtIndex;
        _originalOnOpenStateChanged = Main.panel.statusArea.dateMenu._onOpenStateChanged;
        // Connecting to key of settings
        this.expandMode = this.settings.get_boolean('expand-all');
        settingsHandler = this.settings.connect('changed::expand-all', () => {
            setExpandMode(this.settings.get_boolean('expand-all'));
        });
        // Replace original functions with new ones
        Main.panel.statusArea.dateMenu.menu.connect('open-state-changed', _onOpenStateChanged.bind(Main.panel.statusArea.dateMenu));
        Calendar.NotificationSection.prototype._onNotificationAdded = _onNotificationAdded;
        MessageList.Message.prototype.expand = expand;
        MessageList.Message.prototype.unexpand = unexpand;
        MessageList.MessageListSection.prototype.addMessageAtIndex = addMessageAtIndex;
    }
    disable() {
        // Put everything back
        Main.panel.statusArea.dateMenu.menu.connect('open-state-changed', _originalOnOpenStateChanged.bind(Main.panel.statusArea.dateMenu));
        Calendar.NotificationSection.prototype._onNotificationAdded = _originalOnNotificationAdded;
        MessageList.Message.prototype.expand = originalExpand;
        MessageList.Message.prototype.unexpand = originalUnexpand;
        MessageList.Message.prototype.addMessageAtIndex = originalAddMessageAtIndex;
        this._destroy();
        this._extension = null;
    }
    _destroy() {
        this.settings.disconnect(settingsHandler);
    }
}

function init(meta) {
    _extension = new Extension(meta.uuid);
    return _extension;
}